@@ -350,17 +350,14 @@ static void init_xml_chartypes(void)
     for(i = 0x0483; i <= 0x0486; i++) xml_char_map[i] |= X;
     for(i = 0x0591; i <= 0x05a1; i++) xml_char_map[i] |= X;
     for(i = 0x05a3; i <= 0x05b9; i++) xml_char_map[i] |= X;
-    xml_char_map[0x05bb] |= X;
-    xml_char_map[0x05bd] |= X;
+    for(i = 0x05bb; i <= 0x05bd; i++) xml_char_map[i] |= X;
     xml_char_map[0x05bf] |= X;
     for(i = 0x05c1; i <= 0x05c2; i++) xml_char_map[i] |= X;
     xml_char_map[0x05c4] |= X;
-    xml_char_map[0x064b] |= X;
-    xml_char_map[0x0652] |= X;
+    for(i = 0x064b; i <= 0x0652; i++) xml_char_map[i] |= X;
     xml_char_map[0x0670] |= X;
     for(i = 0x06d6; i <= 0x06dc; i++) xml_char_map[i] |= X;
-    xml_char_map[0x06dd] |= X;
-    xml_char_map[0x06df] |= X;
+    for(i = 0x06dd; i <= 0x06df; i++) xml_char_map[i] |= X;
     for(i = 0x06e0; i <= 0x06e4; i++) xml_char_map[i] |= X;
     for(i = 0x06e7; i <= 0x06e8; i++) xml_char_map[i] |= X;
     for(i = 0x06ea; i <= 0x06ed; i++) xml_char_map[i] |= X;
