@@ -1,7 +1,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
-#ifdef WIN32
+#if defined(WIN32) && ! defined(__CYGWIN__)
 #include <io.h>
 #include <fcntl.h>
 #endif
@@ -9,11 +9,13 @@
 #include "charset.h"
 #include "string16.h"
 #include "dtd.h"
+#include "url.h"
 #include "input.h"
 #include "xmlparser.h"
 #include "stdio16.h"
 #include "version.h"
 #include "namespaces.h"
+#include "infoset-print.h"
 
 int attr_compare(const void *a, const void *b);
 void print_tree(Parser p, XBit bit);
@@ -24,20 +26,26 @@ void print_attrs(ElementDefinition e, Attribute a);
 void print_text(Char *text);
 void print_text_bit(Char *text);
 void dtd_cb(XBit bit, void *arg);
+void dtd_cb2(XBit bit, void *arg);
 void print_canonical_dtd(Parser p, const Char *name);
 InputSource entity_open(Entity ent, void *arg);
+static const char8 *minimal_uri(const char8 *uri, const char8 *base);
 
 int verbose = 0, expand = 1, nsgml = 0,
     attr_defaults = 0, merge = 0, strict_xml = 0, tree = 0, validate = 0,
-    xml_space = 0, namespaces = 0, simple_error = 0, experiment = 0;
-enum {o_unspec, o_none, o_bits, o_plain, o_can1, o_can2, o_can3} output_format = o_unspec;
-char *enc_name = 0;
+    xml_space = 0, namespaces = 0, simple_error = 0, experiment = 0,
+    read_dtd = 0;
+enum {o_unspec, o_none, o_bits, o_plain, o_can1, o_can2, o_can3, o_infoset} output_format = o_unspec;
+char *enc_name = 0, *base_uri = 0;
 CharacterEncoding encoding = CE_unknown;
 InputSource source = 0;
 int need_canonical_dtd = 0;
 
 #define canonical_output (output_format >= o_can1)
 
+Vector(XBit, bits);
+Vector(XBit, dtd_bits);
+
 int main(int argc, char **argv)
 {
     int i;
@@ -74,7 +82,12 @@ int main(int argc, char **argv)
 		output_format = o_bits;
 		break;
 	    case 'o':
-		switch(argv[++i][0])
+		if(++i >= argc)
+		{
+		    fprintf(stderr, "-o requires argument\n");
+		    return 1;
+		}
+		switch(argv[i][0])
 		{
 		case 'b':
 		    output_format = o_bits;
@@ -96,6 +109,12 @@ int main(int argc, char **argv)
 		    output_format = o_can3;
 		    need_canonical_dtd = 1;
 		    break;
+		case 'i':
+		    output_format = o_infoset;
+		    namespaces = 1;
+		    attr_defaults = 1;
+		    merge = 0;
+		    break;
 		default:
 		    fprintf(stderr, "bad output format %s\n", argv[i]);
 		    return 1;
@@ -111,7 +130,12 @@ int main(int argc, char **argv)
 		namespaces = 1;
 		break;
 	    case 'c':
-		enc_name = argv[++i];
+		if(++i >= argc)
+		{
+		    fprintf(stderr, "-c requires argument\n");
+		    return 1;
+		}
+		enc_name = argv[i];
 		break;
 	    case 'm':
 		merge = 1;
@@ -129,12 +153,23 @@ int main(int argc, char **argv)
 	    case 'z':
 		simple_error = 1;
 		break;
+	    case 'u':
+		if(++i >= argc)
+		{
+		    fprintf(stderr, "-u requires argument\n");
+		    return 1;
+		}
+		base_uri = argv[i];
+		break;
+	    case 'd':
+		read_dtd = 1;
+		break;
 	    case '.':
 		experiment = 1;
 		break;
 	    default:
 		fprintf(stderr, 
-			"usage: rxp [-abemnNsStvVx] [-o b|0|1|2|3] [-c encoding] [url]\n");
+			"usage: rxp [-abemnNsStvVx] [-o b|0|1|2|3] [-c encoding] [-u base_uri] [url]\n");
 		return 1;
 	    }
     }
@@ -153,6 +188,14 @@ int main(int argc, char **argv)
     else
 	source = SourceFromStream("<stdin>", stdin);
 
+    if(base_uri)
+    {
+	/* Merge with default base URI so we can use a filename as base */
+	base_uri = url_merge(base_uri, 0, 0, 0, 0, 0);
+	EntitySetBaseURL(source->entity, base_uri);
+	free(base_uri);
+    }
+
     if(!source)
 	return 1;
 
@@ -164,6 +207,12 @@ int main(int argc, char **argv)
     if(validate > 1)
 	ParserSetFlag(p, ErrorOnValidityErrors, 1);
 
+    if(read_dtd)
+    {
+	ParserSetFlag(p, TrustSDD, 0);
+	ParserSetFlag(p, ProcessDTD, 1);
+    }
+
     if(xml_space)
 	ParserSetFlag(p, XMLSpace, 1);
 
@@ -182,6 +231,13 @@ int main(int argc, char **argv)
 	ParserSetCallbackArg(p, p);
     }
 
+    if(output_format == o_infoset)
+    {
+	ParserSetFlag(p, ReturnNamespaceAttributes, 1);
+	ParserSetDtdCallback(p, dtd_cb2);
+	ParserSetCallbackArg(p, p);
+    }
+
     ParserSetFlag(p, SimpleErrorFormat, simple_error);
 
     if(attr_defaults)
@@ -278,11 +334,31 @@ int main(int argc, char **argv)
 	Printf("?>\n");
     }
 
+    VectorInit(bits);
+
     while(1)
     {
 	XBit bit;
 
-	if(tree)
+	if(output_format == o_infoset)
+	{
+	    bit = ReadXTree(p);
+	    if(bit->type == XBIT_dtd)
+	    {
+		bit->children = dtd_bits;
+		bit->nchildren = VectorCount(dtd_bits);
+	    }
+	    if(bit->type == XBIT_error)
+		ParserPerror(p, bit);
+	    else if(bit->type == XBIT_eof)
+	    {
+		infoset_print(Stdout, p, bits, VectorCount(bits));
+		return 0;
+	    }
+	    else
+		VectorPush(bits, bit);
+	}
+	else if(tree)
 	{
 	    bit = ReadXTree(p);
 	    print_tree(p, bit);
@@ -317,7 +393,10 @@ int main(int argc, char **argv)
 	if(tree)
 	    FreeXTree(bit);
 	else
-	    FreeXBit(bit);
+	{
+	    if(output_format != o_infoset)
+		FreeXBit(bit);
+	}
     }
 }
 
@@ -562,6 +641,20 @@ void dtd_cb(XBit bit, void *arg)
 {
     Printf("In DTD: ");
     print_bit(arg, bit);
+    FreeXBit(bit);
+}
+	
+void dtd_cb2(XBit bit, void *arg)
+{
+    XBit copy;
+
+    if(bit->type == XBIT_comment)
+	return;
+
+    copy = malloc(sizeof(*copy));
+
+    *copy = *bit;
+    VectorPush(dtd_bits, copy);
 }
 	
 void print_text(Char *text)
@@ -723,7 +816,8 @@ void print_canonical_dtd(Parser p, const Char *name)
 	    ents[i++] = ent;
     }
 
-    if(nnot + nent == 0)
+    if((output_format != o_can3 && nnot == 0) ||
+       (output_format == o_can3 && nnot + nent == 0))
 	/* Don't produce a DOCTYPE if there's nothing to go in it.
 	(The definition doesn't say this, but that's what the Oasis
 	 test output has.) */
@@ -744,7 +838,7 @@ void print_canonical_dtd(Parser p, const Char *name)
 	not = nots[i];
 
 	if(not->systemid)
-	    uri = NotationURL(not);
+	    uri = minimal_uri(NotationURL(not), EntityURL(p->document_entity));
 	else
 	    uri = 0;
 
@@ -765,7 +859,7 @@ void print_canonical_dtd(Parser p, const Char *name)
 	{
 	    ent = ents[i];
 
-	    uri = EntityURL(ent);
+	    uri = minimal_uri(EntityURL(ent), EntityURL(p->document_entity));
 
 	    Printf("<!ENTITY %S ", ent->name);
 	    if(ent->publicid)
@@ -777,3 +871,31 @@ void print_canonical_dtd(Parser p, const Char *name)
 		
     Printf("]>\n");
 }
+
+/* 
+ * Find minimal URI relative to some base URI.
+ * This is a hack for the benefit of canonical forms, don't rely on it
+ * for anything else.
+ */
+static const char8 *minimal_uri(const char8 *uri, const char8 *base)
+{
+    const char8 *u, *b;
+
+    /* Find common prefix */
+
+    for(u=uri, b=base; *u == *b; u++, b++)
+	;
+
+    /* Go back to a slash */
+
+    while(u >= uri && *u != '/')
+	u--, b--;
+
+    if(*u != '/')
+	return uri;		/* nothing in common */
+
+    if(strchr(b+1, '/'))
+	return uri;		/* too hard */
+
+    return u+1;
+}
