@@ -37,10 +37,13 @@
 #define ERR2(m,x,y) fprintf(stderr,m,x,y)
 #endif
 
-#ifdef WIN32
+#if defined(WIN32) && !defined(__CYGWIN__)
 #undef boolean
 #include <winsock.h>
 #include <fcntl.h>
+#include <io.h>
+#else
+#include <unistd.h>
 #endif
 
 #include "charset.h"
@@ -69,6 +72,9 @@ struct _FILE16 {
     int flags;
     CharacterEncoding enc;
     char16 save;
+    /* There are incount unread bytes starting at inbuf[inoffset] */
+    unsigned char inbuf[4096];
+    int incount, inoffset;
 };
 
 #define FILE16_read             0x01
@@ -76,6 +82,10 @@ struct _FILE16 {
 #define FILE16_close_underlying 0x04
 #define FILE16_crlf		0x08
 #define FILE16_needs_binary     0x8000
+#define FILE16_error		0x4000
+#define FILE16_eof		0x2000
+
+static void filbuf(FILE16 *file);
 
 static int FileRead(FILE16 *file, unsigned char *buf, int max_count);
 static int FileWrite(FILE16 *file, const unsigned char *buf, int count);
@@ -83,13 +93,20 @@ static int FileSeek(FILE16 *file, long offset, int ptrname);
 static int FileClose(FILE16 *file);
 static int FileFlush(FILE16 *file);
 
+static int FDRead(FILE16 *file, unsigned char *buf, int max_count);
+static int FDWrite(FILE16 *file, const unsigned char *buf, int count);
+static int FDSeek(FILE16 *file, long offset, int ptrname);
+static int FDClose(FILE16 *file);
+static int FDFlush(FILE16 *file);
+
 static int StringRead(FILE16 *file, unsigned char *buf, int max_count);
 static int StringWrite(FILE16 *file, const unsigned char *buf, int count);
+static int StringWriteTrunc(FILE16 *file, const unsigned char *buf, int count);
 static int StringSeek(FILE16 *file, long offset, int ptrname);
 static int StringClose(FILE16 *file);
 static int StringFlush(FILE16 *file);
 
-#ifdef WIN32
+#if defined(WIN32) && ! defined(__CYGWIN__)
 #ifdef SOCKETS_IMPLEMENTED
 static int WinsockRead(FILE16 *file, unsigned char *buf, int max_count);
 static int WinsockWrite(FILE16 *file, const unsigned char *buf, int count);
@@ -142,8 +159,22 @@ void deinit_stdio16(void) {
     if(Stderr_open) Fclose(Stderr);
 }
 
-static int ConvertASCII(const char8 *buf, int count, FILE16 *file);
-static int ConvertUTF16(const char16 *buf, int count, FILE16 *file);
+/* Return the size (in bytes) of nul in the given encoding */
+
+static int NullSize(CharacterEncoding enc)
+{
+    switch(enc)
+    {
+    default:
+	return 1;
+
+    case CE_UTF_16B:
+    case CE_ISO_10646_UCS_2B:
+    case CE_UTF_16L:
+    case CE_ISO_10646_UCS_2L:
+	return 2;
+    }
+}
 
 /* Output an ASCII buffer in the specified encoding */
 
@@ -182,10 +213,10 @@ static int ConvertASCII(const char8 *buf, int count, FILE16 *file)
 		    outbuf[j++] = '\r';
 		outbuf[j++] = c;
 	    }
-	    return file->write(file, outbuf, j);
+	    return Writeu(file, outbuf, j);
 	}
 	else
-	    return file->write(file, (unsigned char *)buf, count);
+	    return Writeu(file, (unsigned char *)buf, count);
 
     case CE_UTF_8:
 	for(i=j=0; i<count; i++)
@@ -201,7 +232,7 @@ static int ConvertASCII(const char8 *buf, int count, FILE16 *file)
 		outbuf[j++] = 0x80 + (c & 0x3f);
 	    }
 	}
-	return file->write(file, outbuf, j);
+	return Writeu(file, outbuf, j);
 
     case CE_UTF_16B:
     case CE_ISO_10646_UCS_2B:
@@ -216,7 +247,7 @@ static int ConvertASCII(const char8 *buf, int count, FILE16 *file)
 	    outbuf[j++] = 0;
 	    outbuf[j++] = c;
 	}
-	return file->write(file, outbuf, j);
+	return Writeu(file, outbuf, j);
 
     case CE_UTF_16L:
     case CE_ISO_10646_UCS_2L:
@@ -231,7 +262,7 @@ static int ConvertASCII(const char8 *buf, int count, FILE16 *file)
 	    outbuf[j++] = c;
 	    outbuf[j++] = 0;
 	}
-	return file->write(file, outbuf, j);
+	return Writeu(file, outbuf, j);
 
     default:
       ERR2("Bad output character encoding %d (%s)\n",
@@ -269,7 +300,7 @@ static int ConvertUTF16(const char16 *buf, int count, FILE16 *file)
 	    else
 		outbuf[j++] = '?';
 	}
-	return file->write(file, outbuf, j);
+	return Writeu(file, outbuf, j);
 
     case CE_ISO_8859_2:
     case CE_ISO_8859_3:
@@ -292,7 +323,7 @@ static int ConvertUTF16(const char16 *buf, int count, FILE16 *file)
 	    else
 		outbuf[j++] = '?';
 	}
-	return file->write(file, outbuf, j);
+	return Writeu(file, outbuf, j);
 
     case CE_UTF_8:
 	for(i=j=0; i<count; i++)
@@ -325,7 +356,7 @@ static int ConvertUTF16(const char16 *buf, int count, FILE16 *file)
 		outbuf[j++] = 0x80 + (c & 0x3f);
 	    }
 	}
-	return file->write(file, outbuf, j);
+	return Writeu(file, outbuf, j);
 
     case CE_UTF_16B:
     case CE_ISO_10646_UCS_2B:
@@ -341,7 +372,7 @@ static int ConvertUTF16(const char16 *buf, int count, FILE16 *file)
 	    outbuf[j++] = (c & 0xff);
 
 	}
-	return file->write(file, outbuf, j);
+	return Writeu(file, outbuf, j);
 
     case CE_UTF_16L:
     case CE_ISO_10646_UCS_2L:
@@ -357,7 +388,7 @@ static int ConvertUTF16(const char16 *buf, int count, FILE16 *file)
 	    outbuf[j++] = (c >> 8);
 
 	}
-	return file->write(file, outbuf, j);
+	return Writeu(file, outbuf, j);
 
     default:
       ERR2("Bad output character encoding %d (%s)\n",
@@ -369,14 +400,22 @@ static int ConvertUTF16(const char16 *buf, int count, FILE16 *file)
     }
 }
 
+#if 0
 int Readu(FILE16 *file, unsigned char *buf, int max_count)
 {
     return file->read(file, buf, max_count);
 }
+#endif
 
 int Writeu(FILE16 *file, unsigned char *buf, int count)
 {
-    return file->write(file, buf, count);
+    int ret;
+
+    ret = file->write(file, buf, count);
+    if(ret < 0)
+	file->flags |= FILE16_error;
+
+    return ret;
 }
 
 int Fclose(FILE16 *file)
@@ -398,6 +437,8 @@ int Fclose(FILE16 *file)
 
 int Fseek(FILE16 *file, long offset, int ptrname)
 {
+    file->incount = file->inoffset = 0;
+    file->flags &= ~(FILE16_error | FILE16_eof);
     return file->seek(file, offset, ptrname);
 }
 
@@ -406,6 +447,70 @@ int Fflush(FILE16 *file)
     return file->flush(file);
 }
 
+int Ferror(FILE16 *file)
+{
+    return file->flags & FILE16_error;
+}
+
+int Feof(FILE16 *file)
+{
+    return file->flags & FILE16_eof;
+}
+
+int Getu(FILE16 *file)
+{
+    filbuf(file);
+    if(file->flags & (FILE16_error | FILE16_eof))
+	return EOF;
+    file->incount--;
+    return file->inbuf[file->inoffset++];
+}
+
+int Readu(FILE16 *file, unsigned char *buf, int max_count)
+{
+    int count=0, lump;
+
+    while(count < max_count)
+    {
+	filbuf(file);
+	if(file->flags & FILE16_error)
+	    return 0;
+	if(file->flags & FILE16_eof)
+	    break;
+	lump = max_count - count;
+	if(lump > file->incount)
+	    lump = file->incount;
+	memcpy(buf+count, file->inbuf+file->inoffset, lump);
+	count += lump;
+	file->inoffset += lump;
+	file->incount -= lump;
+    }
+    
+    return count;
+}
+
+/* After calling filbuf, either file->incount > 0 or file->flags has
+   FILE16_eof or FILE16_error set. */
+
+static void filbuf(FILE16 *file)
+{
+    int ret;
+
+    if(file->incount > 0)
+	return;
+
+    ret = file->read(file, file->inbuf, sizeof(file->inbuf));
+    if(ret < 0)
+	file->flags |= FILE16_error;
+    else if(ret == 0)
+	file->flags |= FILE16_eof;
+    else
+    {
+	file->inoffset = 0;
+	file->incount = ret;
+    }
+}
+
 FILE *GetFILE(FILE16 *file)
 {
     if(file->read == FileRead)
@@ -432,6 +537,16 @@ CharacterEncoding GetFileEncoding(FILE16 *file)
     return file->enc;
 }
 
+void SetNormalizeLineEnd(FILE16 *file, int nle)
+{
+#if defined(WIN32)
+    if(nle)
+	file->flags |= FILE16_crlf;
+    else
+	file->flags &= ~FILE16_crlf;
+#endif
+}
+
 int Fprintf(FILE16 *file, const char *format, ...)
 {
     va_list args;
@@ -453,6 +568,14 @@ int Sprintf(void *buf, CharacterEncoding enc, const char *format, ...)
     return Vsprintf(buf, enc, format, args);
 }
 
+int Snprintf(void *buf, size_t size, CharacterEncoding enc, 
+	     const char *format, ...)
+{
+    va_list args;
+    va_start(args, format);
+    return Vsnprintf(buf, size, enc, format, args);
+}
+
 int Vprintf(const char *format, va_list args)
 {
     return Vfprintf(Stdout, format, args);
@@ -462,17 +585,34 @@ int Vsprintf(void *buf, CharacterEncoding enc, const char *format,
 	     va_list args)
 {
     int nchars;
-    FILE16 file = {0, 0, -1, StringRead, StringWrite, StringSeek, StringFlush, StringClose, FILE16_write};
+    FILE16 file = {0, 0, -1, 0, StringWrite, 0, StringFlush, StringClose, FILE16_write};
 
     file.handle = buf;
     file.enc = enc;
 
     nchars = Vfprintf(&file, format, args);
-    file.close(&file);		/* Fclose would try to free it */
+    file.close(&file);		/* Fclose would try to free file */
 
     return nchars;
 }
 
+int Vsnprintf(void *buf, size_t size, CharacterEncoding enc,
+	      const char *format, va_list args)
+{
+    int nchars;
+    FILE16 file = {0, 0, -1, 0, StringWriteTrunc, 0, StringFlush, StringClose, FILE16_write};
+
+    file.handle = buf;
+    file.enc = enc;
+    file.handle3 = size - NullSize(enc); /* make sure we can null-terminate */
+
+    nchars = Vfprintf(&file, format, args);
+    file.handle3 = size;	/* ready for null-termination */
+    file.close(&file);		/* Fclose would try to free file */
+    
+    return nchars;
+}
+
 #define put(x) {nchars++; if(count == sizeof(buf)) {if(ConvertASCII(buf, count, file) == -1) return -1; count = 0;} buf[count++] = x;}
 
 int Vfprintf(FILE16 *file, const char *format, va_list args)
@@ -718,15 +858,87 @@ static FILE16 *MakeFILE16(const char *type)
 
     file->flags = 0;
     if(*type == 'r')
+    {
 	file->flags |= FILE16_read;
-    else
+	type++;
+    }
+    if(*type == 'w')
 	file->flags |= FILE16_write;
 
     file->enc = InternalCharacterEncoding;
 
+    file->incount = file->inoffset = 0;
+
     return file;
 }
 
+FILE16 *MakeFILE16FromFD(int fd, const char *type)
+{
+    FILE16 *file;
+
+    if(!(file = MakeFILE16(type)))
+	return 0;
+
+#if defined(WIN32)
+    file->flags |= FILE16_crlf;
+#endif
+
+    file->read = FDRead;
+    file->write = FDWrite;
+    file->seek = FDSeek;
+    file->close = FDClose;
+    file->flush = FDFlush;
+    file->handle2 = fd;
+
+    return file;
+}
+
+static int FDRead(FILE16 *file, unsigned char *buf, int max_count)
+{
+    int fd = file->handle2;
+    int count = 0;
+
+    count = read(fd, buf, max_count);
+
+    return count;
+}
+
+static int FDWrite(FILE16 *file, const unsigned char *buf, int count)
+{
+    int fd = file->handle2;
+    int ret;
+
+    while(count > 0)
+    {
+	ret = write(fd, buf, count);
+	if(ret < 0)
+	    return ret;
+	count -= ret;
+	buf += ret;
+    }
+
+    return 0;
+}
+
+static int FDSeek(FILE16 *file, long offset, int ptrname)
+{
+    int fd = file->handle2;
+
+    return lseek(fd, offset, ptrname);
+}
+
+static int FDClose(FILE16 *file)
+{
+    int fd = file->handle2;
+
+    return (file->flags & FILE16_close_underlying) ? close(fd) : 0;
+}
+
+static int FDFlush(FILE16 *file)
+{
+    return 0;
+}
+
 FILE16 *MakeFILE16FromFILE(FILE *f, const char *type)
 {
     FILE16 *file;
@@ -734,7 +946,7 @@ FILE16 *MakeFILE16FromFILE(FILE *f, const char *type)
     if(!(file = MakeFILE16(type)))
 	return 0;
 
-#ifdef WIN32
+#if defined(WIN32)
     /* We need to do Windows-style lf <-> crlf conversion ourselves, 
        because the standard i/o library doesn't know about 16-bit 
        characters. We don't want it to insert cr before output bytes 
@@ -767,14 +979,25 @@ static int FileRead(FILE16 *file, unsigned char *buf, int max_count)
     FILE *f = file->handle;
     int count = 0;
 
-#ifdef WIN32
+#if defined(WIN32)
     if(file->flags & FILE16_needs_binary)
     {
+#ifdef __CYGWIN__
+	setmode(fileno(f), _O_BINARY);
+#else
 	_setmode(fileno(f), _O_BINARY);
+#endif
 	file->flags &= ~FILE16_needs_binary;
     }
 #endif
 
+    /* Terminal EOF is sticky for fread() on Linux, even though feof() is true.
+       How can they have got this wrong and never noticed it?
+       So check feof() here.
+    */
+    if(feof(f))
+	return 0;
+
     count = fread(buf, 1, max_count, f);
 
     return ferror(f) ? -1 : count;
@@ -784,10 +1007,14 @@ static int FileWrite(FILE16 *file, const unsigned char *buf, int count)
 {
     FILE *f = file->handle;
 
-#ifdef WIN32
+#if defined(WIN32)
     if(file->flags & FILE16_needs_binary)
     {
+#ifdef __CYGWIN__
+	setmode(fileno(f), _O_BINARY);
+#else
 	_setmode(fileno(f), _O_BINARY);
+#endif
 	file->flags &= ~FILE16_needs_binary;
     }
 #endif
@@ -867,6 +1094,25 @@ static int StringWrite(FILE16 *file, const unsigned char *buf, int count)
     return 0;
 }
 
+/* Like StringWrite, but ignores overflow rather than returning an error.
+   Used for Vsnprintf. */
+
+static int StringWriteTrunc(FILE16 *file, const unsigned char *buf, int count)
+{
+    char *p = (char *)file->handle + file->handle2;
+
+    if(file->handle3 >= 0 && file->handle2 + count > file->handle3)
+	/* XXX This doesn't really work; a character might be truncated.
+	   Safe for fixed-size encodings if buffer size is a multiple
+	   of the character size. */
+	count = file->handle3 - file->handle2;
+
+    memcpy(p, buf, count);
+    file->handle2 += count;
+
+    return 0;
+}
+
 static int StringSeek(FILE16 *file, long offset, int ptrname)
 {
     switch(ptrname)
@@ -908,7 +1154,7 @@ static int StringFlush(FILE16 *file)
 }
 
 
-#ifdef WIN32
+#if defined(WIN32) && ! defined(__CYGWIN__)
 #ifdef SOCKETS_IMPLEMENTED
 
 FILE16 *MakeFILE16FromWinsock(int sock, const char *type)
@@ -932,7 +1178,7 @@ FILE16 *MakeFILE16FromWinsock(int sock, const char *type)
 
 static int WinsockRead(FILE16 *file, unsigned char *buf, int max_count)
 {
-    int f = (int)file->handle2;
+    int f = file->handle2;
     int count;
 
     /* "Relax" said the nightman, we are programmed to recv() */
@@ -943,9 +1189,9 @@ static int WinsockRead(FILE16 *file, unsigned char *buf, int max_count)
 
 static int WinsockWrite(FILE16 *file, const unsigned char *buf, int count)
 {
-    /* Not yet implemented */
+    int f = file->handle2;
 
-    return -1;
+    return send(f, buf, count ,0);
 }
 
 static int WinsockSeek(FILE16 *file, long offset, int ptrname)
