@@ -49,28 +49,16 @@
 extern FILE *fdopen(int, const char *);
 #endif
 
-#ifdef WIN32
+#if defined(WIN32) && ! defined(__CYGWIN__)
 #include <direct.h>
-#endif
-
-#ifdef SOCKETS_IMPLEMENTED
-
-#ifdef WIN32
-#undef boolean
-#include <winsock.h>
-#include <fcntl.h>
 #else
 #include <unistd.h>
-#include <netdb.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#endif
-
 #endif
 
 #include "string16.h"
 #include "stdio16.h"
 #include "url.h"
+#include "http.h"
 
 #ifdef HAVE_LIBZ
 #include "zlib.h"
@@ -80,12 +68,9 @@ extern FILE *fdopen(int, const char *);
 #endif
 #endif
 
-static FILE16 *http_open(const char *url, 
-		       const char *host, int port, const char *path,
-		       const char *type);
 static FILE16 *file_open(const char *url,
 		       const char *host, int port, const char *path,
-		       const char *type);
+		       const char *type, char **redirected_url);
 
 static void parse_url(const char *url, 
 		      char **scheme, char **host, int *port, char **path);
@@ -94,70 +79,22 @@ static void parse_url(const char *url,
 
 struct {
     char *scheme; 
-    FILE16 *(*open)(const char *, const char *, int, const char *, const char *);
+    FILE16 *(*open)(const char *, const char *, int, const char *, const char *, char **);
 } schemes[] = {
     {(char *)"http", http_open},
     {(char *)"file", file_open},
 };
 #define NSCHEME (sizeof(schemes) / sizeof(schemes[0]))
 
-static int proxy_port;
-static char *proxy_host = 0;
-
 int init_url(void)
 {
-    char *p;
 
-#ifdef WIN32
-    /* random magic for MS Windows */
-
-    WORD version = MAKEWORD(1, 1);
-    WSADATA wsaData;
-    int err = WSAStartup(version, &wsaData);
-    if (err)
-    {
-	LT_ERROR(LEFILE, "Error: can't init HTTP interface\n");
-	return -1;
-    }
-    else if(LOBYTE(wsaData.wVersion) != 1 || HIBYTE(wsaData.wVersion) != 1)
-    {
-	LT_ERROR(LEFILE, "Error: wrong version of WINSOCK\n");
-	WSACleanup();
-	return -1;
-    }
-#endif
-
-    /* See if we are supposed to use a proxy */
-
-    if((p = getenv("http_proxy")))
-    {
-	char *colon, *slash;
-	if(strncmp(p, "http://", 7) == 0)
-	    p += 7;
-	proxy_host = strdup8(p);
-	if((slash = strchr(proxy_host, '/')))
-	    *slash = 0;
-	if((colon = strchr(proxy_host, ':')))
-	{
-	    proxy_port = atoi(colon+1);
-	    *colon = 0;
-	}
-	else
-	    proxy_port = 80;
-    }
-
-    return 0;
+    return init_http();
 }
 
 void deinit_url(void)
 {
-#ifdef WIN32
-    /* No doubt there's some Windozy nonsense that should go here,
-       but I have no idea what it is. */
-#endif
-
-    if(proxy_host)
-	Free(proxy_host);
+    deinit_http();
 }
 
 /* Construct a default base URL, essentially file://`pwd`/ */
@@ -176,7 +113,7 @@ char *default_base_url(void)
 
     len = strlen(buf);
 
-#ifdef WIN32
+#if defined(WIN32) && ! defined(__CYGWIN__)
 
     /* DOS: translate C:\a\b to file:///C:/a/b/ */
     /* XXX should we escape anything? */
@@ -432,7 +369,8 @@ bad:
  * is null, file://`pwd`/ is used, which is the right thing to do for
  * filenames.  If base is "", there is no base URL and relative
  * URLs will fail.
- * If merged_url is non-null the resulting URL is stored in it.
+ * If merged_url is non-null the resulting URL (after any scheme-specific
+ * redirection) is stored in it.
  * If type begins "r", the URL is opened for reading, if "w" for
  * writing.  Writing is only supported for file URLs.
  * If the type begins "rl", the data will be copied to a temporary
@@ -443,7 +381,7 @@ bad:
 FILE16 *url_open(const char *url, const char *base, const char *type,
 		 char **merged_url)
 {
-    char *scheme, *host, *path, *m_url;
+    char *scheme, *host, *path, *m_url, *r_url;
     int port, i;
     FILE16 *f;
 #ifdef HAVE_LIBZ
@@ -471,7 +409,7 @@ FILE16 *url_open(const char *url, const char *base, const char *type,
     for(i=0; i<NSCHEME; i++)
 	if(strcmp(scheme, schemes[i].scheme) == 0) 
 	{
-	    f = schemes[i].open(m_url, host, port, path, type);
+	    f = schemes[i].open(m_url, host, port, path, type, &r_url);
 	
 	    Free(scheme);
 	    if(host)
@@ -507,6 +445,11 @@ FILE16 *url_open(const char *url, const char *base, const char *type,
 		 f = MakeFILE16FromGzip(gfile, type);
 	    }
 #endif
+	    if(r_url)
+	    {
+		Free(m_url);
+		m_url = r_url;
+	    }
 	    if(f && merged_url)
 		*merged_url = m_url;
 	    else
@@ -528,243 +471,11 @@ FILE16 *url_open(const char *url, const char *base, const char *type,
     return 0;
 }
 
-/* Open an http URL */
-
-static FILE16 *http_open(const char *url,
-			 const char *host, int port, const char *path,
-			 const char *type)
-{
-#ifndef SOCKETS_IMPLEMENTED
-    LT_ERROR(NEUNSUP, 
-		"http: URLs are not yet implemented on this platform\n");
-    return 0;
-#else
-    FILE16 *f16;
-    struct sockaddr_in addr;
-    struct hostent *hostent;
-    int s, server_major, server_minor, status, count, c;
-    char reason[81];
-    const char *server_host, *request_uri;
-    int server_port;
-#ifndef WIN32
-    FILE *fin,*fout;
-#else
-    int i;
-    char buf[1024];
-#endif
-
-    if(*type != 'r')
-    {
-	LT_ERROR1(LEFILE, "Error: can't open http URL \"%s\" for writing\n",
-		     url);
-	return 0;
-    }
-
-    if(!host)
-    {
-	LT_ERROR1(LEFILE, "Error: no host part in http URL \"%s\"\n", url);
-	return 0;
-    }
-
-    if(proxy_host)
-    {
-	server_host = proxy_host;
-	server_port = proxy_port;
-	request_uri = url;
-    }
-    else
-    {
-	server_host = host;
-	server_port = port == -1 ? 80 : port;
-	request_uri = path;
-    }
-
-    /* Create the socket */
-
-    s = socket(PF_INET, SOCK_STREAM, 0);
-#ifdef WIN32
-    if (s == INVALID_SOCKET) {
-      LT_ERROR1(LEFILE, "Error: system call socket failed: %d\n",
-		   WSAGetLastError());
-    };
-#else
-    if(s == -1) {
-	LT_ERROR1(LEFILE, "Error: system call socket failed: %s\n",
-		     Strerror());
-	return 0;
-    };
-#endif
-
-    /* Find the server address */
-
-    hostent = gethostbyname(server_host);
-    if(!hostent)
-    {
-	LT_ERROR3(LEFILE, "Error: can't find address for %shost \"%s\" "
-		          "in http URL \"%s\"\n",
-		  proxy_host ? "proxy " : "", server_host, url);
-	return 0;
-    }
-
-    memset(&addr, 0, sizeof(addr));
-    addr.sin_family = AF_INET;
-    /* If we were really enthusiastic, we would try all the host's addresses */
-    memcpy(&addr.sin_addr, hostent->h_addr, hostent->h_length);
-    addr.sin_port = htons((unsigned short)server_port);
-
-    /* Connect */
-
-    if(connect(s, (struct sockaddr *)&addr, sizeof(addr)) == -1)
-    {
-	LT_ERROR4(LEFILE, "Error: connection to %shost \"%s\" failed "
-		          "for http URL \"%s\": %s\n",
-		  proxy_host ? "proxy " : "", server_host, url, Strerror());
-	return 0;
-    }
-
-#ifndef WIN32
-#ifdef macintosh
-    fin = fdopen(s, "rb");
-    setvbuf(fin, 0, _IONBF, 0);
-    fout = fdopen(dup(s), "wb");
-#else
-    fin = fdopen(s, "r");
-    setvbuf(fin, 0, _IONBF, 0);
-    fout = fdopen(dup(s), "w");
-#endif
-#endif
-
-    /* Send the request */
-
-    /* 
-     * Apparently on the Macintosh, \n might not be ASCII LF, so we'll
-     * use numerics to be sure.
-     */
-
-#ifdef WIN32
-    sprintf(buf, "GET %s HTTP/1.0\015\012Connection: close\015\012",
-	    request_uri);
-    if (send(s,buf,strlen8(buf),0)==SOCKET_ERROR) {
-      LT_ERROR1(LEFILE, "Error: system call socket failed: %d\n",
-		   WSAGetLastError());
-      /* XXX close the socket? */
-      return 0;
-    };      
-    if(port != -1)
-	sprintf(buf, "Host: %s:%d\015\012\015\012", host, port);
-    else
-	sprintf(buf, "Host: %s\015\012\015\012", host);
-    if (send(s,buf,strlen8(buf),0)==SOCKET_ERROR) {
-      LT_ERROR1(LEFILE, "Error: system call socket failed: %d\n",
-		   WSAGetLastError());
-      /* XXX close the socket? */
-      return 0;
-    };      
-#else
-    fprintf(fout, "GET %s HTTP/1.0\015\012Connection: close\015\012",
-	    request_uri);
-    if(port != -1)
-	fprintf(fout, "Host: %s:%d\015\012\015\012", host, port);
-    else
-	fprintf(fout, "Host: %s\015\012\015\012", host);
-
-    /* We used to test for errors after doing fclose, but this seemed
-       to produce spurious errors under Linux (RedHat 4.2), so now we
-       do fflush and test after that. */
-
-    fflush(fout);
-    if(ferror(fout))
-    {
-	LT_ERROR1(LEWRTF, "Error: write to socket failed: %s\n",Strerror());
-	fclose(fout);
-	fclose(fin);
-	return 0;
-    }
-    fclose(fout);
-#endif
-
-    /* Read the status line */
-#ifdef WIN32
-    for(i=0; i<sizeof(buf)-1; i++)
-    {
-	if(recv(s, &buf[i], 1, 0) != 1)
-	    LT_ERROR1(LEFILE,
-			 "Error: recv error from server for URL \"%s\"\n",
-			 url);
-	if(buf[i] == '\n')
-	    break;
-    }
-    count=sscanf(buf, "HTTP/%d.%d %d %80[^\012]", 
-		 &server_major, &server_minor, &status, reason);
-#else    
-    count=fscanf(fin, "HTTP/%d.%d %d %80[^\012]", 
-		 &server_major, &server_minor, &status, reason);
-#endif
-
-    if(count != 4)
-    {
-	LT_ERROR3(LEFILE,
-		     "Error: bad header from server for URL \"%s\"\n%d %s\n",
-		     url, count, Strerror());
-#ifndef WIN32
-	fclose(fin);
-#endif
-	return 0;
-    }
-
-    if(status != 200)
-    {
-	/* We should handle 301 (redirection) but we don't */
-	LT_ERROR3(LEFILE, "Error: can't retrieve \"%s\": %d %s\n",
-		     url, status, reason);
-#ifndef WIN32
-	fclose(fin);
-#endif
-	return 0;
-    }
-
-    /* Skip other headers */
-
-    count = 0;
-#ifdef WIN32
-    while(recv(s, buf, 1, 0) == 1 && (c = buf[0], 1) || (c = EOF, 0))
-#else
-    while((c = getc(fin)) != EOF)
-#endif
-    {
-	if(c == '\012')
-	    count++;
-	else if(c != '\015')
-	    count = 0;
-	if(count == 2)
-	    break;
-    }
-
-    if(c == EOF)
-    {
-	LT_ERROR1(LEFILE, "Error: EOF in headers retrieving \"%s\"\n", url);
-#ifndef WIN32
-	fclose(fin);
-#endif
-	return 0;
-    }
-
-#ifdef WIN32
-    f16 = MakeFILE16FromWinsock(s, type);
-#else
-    f16 = MakeFILE16FromFILE(fin, type);
-#endif
-
-    SetCloseUnderlying(f16, 1);
-    return f16;
-#endif /* SOCKETS_IMPLEMENTED */
-}
-
 /* Open a file URL (easy, at least on unix) */
 
 static FILE16 *file_open(const char *url,
 			 const char *host, int port, const char *path, 
-			 const char *type)
+			 const char *type, char **redirected_url)
 {
     FILE *f;
     FILE16 *f16;
@@ -773,7 +484,7 @@ static FILE16 *file_open(const char *url,
     if(host && host[0])
 	WARN1(LEFILE, "Warning: ignoring host part in file URL \"%s\"\n", url);
 
-#ifdef WIN32
+#if defined(WIN32) && ! defined(__CYGWIN__)
 
     /* DOS: translate /C:/a/b.c to C:\a\b.c */
 
@@ -841,6 +552,9 @@ static FILE16 *file_open(const char *url,
     f16 = MakeFILE16FromFILE(f, type);
     SetCloseUnderlying(f16, 1);
 
+    if(redirected_url)
+	*redirected_url = 0;
+
     return f16;
 }
 
